/*
 * CELParser.java
 *
 * Created on May 16, 2006, 9:06 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package MBR;
import java.io.BufferedInputStream;     
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.awt.image.Raster;
import java.util.Properties;
import java.text.NumberFormat;
import java.util.Vector;
/**
 *
 * @author jssong
 */
public class CELParser {
    
    /** The name of the CEL file */
    private String fileName;
   
    /** The minimum pixel value encountered in this image. */
    private short minimum=24;

    /** The maximum pixel value encountered in this image. */
    private short maximum=3000;

    /** The position of the data */
    private long dataPos=0;

    private int nColumns, nRows, nCells;
    private int cellMargin;
    private long nOutlierCells, nMaskedCells;
    private int[] outlierPos = new int[2]; // outlierPos[0] = location in file nOutliers
                                           // outlierPos[1] = location in file for list of outliers
    private int nSubGrids; 
    /** RandomAccessFile to read the CEL file. */
    private RandomAccessFile file;
    private static int CEL_INDICATOR = 64;  // 
    private static int VERSION_INDICATOR = 4;  // 
    
    private boolean littleEndian = true;
    
    private float[][] intensities;
    private float[][] std;
    private short[][] pixelCount;
    private byte[] data;
    private float maxIntensity= 0; 
    
    public CELParser (String fileName, int option) {
	this.fileName = fileName;
        try{
            file = new RandomAccessFile(fileName, "r");
            switch(option){
                /* Read CEL files for display */
                case 0:   readInfo();
                    break;
                /* Write new CEL files with bubbles removed */
                case 1:   readHeader();
                    break;
            }
        }
        catch (Exception e) {
        }
    }

    
    
    
    /** Read a string of specified length from the binary file **/
    String getString(int length) throws IOException {
        byte[] buf = new byte[length];
        byte[] sequence_tag = new byte[8];
        
        
        int pos_subtag = 0;
        if(length==0) return "";
        else {
            int pos = 0;
            while (pos<(length)) {
                    int count = file.read(buf, pos, length-pos);
                    pos += count;
            }
        }
        
        dataPos += length;
        return new String(buf);
    }

    /** Read a byte of data **/
    int getByte() throws IOException {
            int b = file.read();
            // return -1 if EOF;
            if(b==-1) return -1;
            ++dataPos;
            return b;
    }

    /** Read 2 bytes and return a short int **/

    int getShort() throws IOException {
            int b0 = getByte();
            if(b0==-1) return -1;
            int b1 = getByte();
            if (littleEndian)
                    return ((b1 << 8) + b0);
            else
                    return ((b0 << 8) + b1);
    }

     int getShort(byte[] b, int offset) {
            if (littleEndian)
                    return (((b[1+offset] & 0xff) << 8) + (b[offset]&0xff));
            else
                    return (((b[0+offset]& 0xff) << 8) + (b[1+offset]& 0xff));
    }
    /** Read 4 bytes and return a signed int **/
    final int getInt() throws IOException {
            int b0 = getByte();
            int b1 = getByte();
            int b2 = getByte();
            int b3 = getByte();
            if (littleEndian)
                    return ((b3<<24) + (b2<<16) + (b1<<8) + b0);
            else
                    return ((b0<<24) + (b1<<16) + (b2<<8) + b3);
    }

    /** Read 4 bytes and return a long int to represent an unsigned int **/
    final long getLongInt() throws IOException {
            long b0 = (long) getByte();
            long b1 = (long) getByte();
            long b2 = (long) getByte();
            long b3 = (long) getByte();
            if (littleEndian)
                    return ((b3<<24) + (b2<<16) + (b1<<8) + b0);
            else
                    return ((b0<<24) + (b1<<16) + (b2<<8) + b3);
    }

    /** Read 4 bytes and return a float **/
    final float getFloat(byte[] b, int offset)  {
            
            int tmp;
            if (littleEndian)
                tmp = (((b[3+offset] & 0xff)<<24) + ( (b[2+offset] & 0xff)<<16) + ((b[1+offset]&0xff)<<8) + (b[offset]&0xff));
            else
                tmp =(((b[offset]& 0xff)<<24) + ((b[1+offset]& 0xff)<<16) + ((b[2+offset]& 0xff)<<8) + (b[3+offset]& 0xff));
            return Float.intBitsToFloat(tmp);
    }
    /****************************************************************************************/

    int getIntValue(int length) throws IOException {
        if(length == 2) {
            int p = getShort();
            return p;
        } else if( length ==4) {
            int p = getInt();
            return p;
        }else return -1;
    }


    
    /****************************************************************************************/
    public void readInfo() throws IOException {
       // while(skipCount >0) skipCount -= file.skipBytes(skipCount);

        int MAGIC_NUMBER = getInt();
        int VERSION_NUMBER = getInt();
        String header;
        String algorithm;
        String algorithm_param;
        if(MAGIC_NUMBER != CEL_INDICATOR || VERSION_NUMBER != VERSION_INDICATOR){
            file.close();
            System.out.println("NOT A RECOGNIZABLE CEL FILE.  IT HAS TO BE VERSION 4.");
        } else{
            nColumns = getInt();
            nRows = getInt();
            nCells = getInt();
            
            intensities = new float[nRows][nColumns];
            std = new float[nRows][nColumns];
            pixelCount = new short[nRows][nColumns];
            
            int stringlength = getInt();
            //System.out.println(nColumns+" " + nRows + " " + nCells);
            header = getString(stringlength);
            //System.out.println(header);
            
            stringlength = getInt();
            algorithm = getString(stringlength);
            
            stringlength = getInt();
            algorithm_param = getString(stringlength);
            cellMargin = getInt();
            outlierPos[0] = (int) dataPos;
            //System.out.println("Cell Margin = " + cellMargin);
            
            nOutlierCells = getLongInt();
            nMaskedCells = getLongInt();
            nSubGrids = getInt();
            //System.out.println(nOutlierCells + " " + nMaskedCells + " " + nSubGrids);
            
            outlierPos[1] = (int)(dataPos + (long)(nCells*10) + (long)(nMaskedCells*4));
            
            int datalength = nCells*10;
            byte[] buf = new byte[datalength]; // 2 floats + 1 short
                
            int pos = 0;
            while (pos<(datalength)) {
                    int count = file.read(buf, pos, datalength-pos);
                    pos += count;
            }
            dataPos += datalength;
            
            parseData(buf);
            file.close();
            
        }      
    }

    /*** Read header information for rewriting ***/
    public void readHeader() throws IOException {
       
        int MAGIC_NUMBER = getInt();
        int VERSION_NUMBER = getInt();
        String header;
        String algorithm;
        String algorithm_param;
        if(MAGIC_NUMBER != CEL_INDICATOR || VERSION_NUMBER != VERSION_INDICATOR){
            file.close();
            System.out.println("NOT A RECOGNIZABLE CEL FILE.  IT HAS TO BE VERSION 4.");
        } else{
            nColumns = getInt();
            nRows = getInt();
            nCells = getInt();
            
            int stringlength = getInt();
            //System.out.println(nColumns+" " + nRows + " " + nCells);
            header = getString(stringlength);
            //System.out.println(header);
            
            stringlength = getInt();
            algorithm = getString(stringlength);
            
            stringlength = getInt();
            algorithm_param = getString(stringlength);
            cellMargin = getInt();
            
            outlierPos[0] = (int) dataPos;
            nOutlierCells = getLongInt();
            nMaskedCells = getLongInt();
            nSubGrids = getInt();
            
            outlierPos[1] = (int)(dataPos + (long)(nCells*10) + (long)(nMaskedCells*4));
            file.close();

        }      
    }
    private void parseData(byte[] data){
        int offset = 0;

        for (int j = 0; j < nColumns; j ++){
            for (int i =0; i < nRows; i ++){
                intensities[i][j] =  getFloat(data,offset);
                //if (intensities[i][j] > maxIntensity) maxIntensity = intensities[i][j];
                offset += 4;
                std[i][j] =  getFloat(data,offset);
                offset += 4;
                pixelCount[i][j] = (short) getShort(data,offset);
                offset += 2;
            }
        }
    }
    
    public String writeNewCEL(Vector outliers){
        try{
            file = new RandomAccessFile(fileName, "r");
            int filelength = (int) file.length();
            byte[] all = new byte[filelength];
            int pos = 0;
            int newOutliers = outliers.size();
            while (pos<(filelength)) {
                    int count = file.read(all, pos, filelength-pos);
                    pos += count;
            }
            file.close();
            
            int dot = fileName.lastIndexOf(".");
            String fo = fileName.substring(0,dot);
            fo += "_MBR.CEL";
            java.io.BufferedOutputStream writer = new java.io.BufferedOutputStream(new java.io.FileOutputStream(fo));
            
            writer.write(all,0,outlierPos[0]);
            writer.write(int2byte(outliers.size()));
            writer.write(all, outlierPos[0]+4, outlierPos[1]- outlierPos[0]-4);
            for (int i = 0; i < outliers.size(); i ++){
                writer.write( short2byte((short[]) outliers.elementAt(i)));
            }
            
            pos = outlierPos[1] + (int)nOutlierCells*4;
            writer.write(all,pos,  filelength - pos);
            writer.close();
            return fo;
        } catch (IOException e) {
            return "";
        }
    }
    public short[][] readOutliers(){
        short[][] outlierlist = new short[(int)nOutlierCells][2];
        try{
            file = new RandomAccessFile(fileName, "r");
           
            int filelength = (int) file.length();
            byte[] all = new byte[filelength];
            int pos = 0;
            while (pos<(filelength)) {
                    int count = file.read(all, pos, filelength-pos);
                    pos += count;
            }
            file.close();
            
            int upperlimit = outlierPos[1] + (int) nOutlierCells*4;
            byte[] tmp = new byte[2];
            int n = 0;
            for (int i =outlierPos[1]  ; i<upperlimit; i= i+4){
                tmp[0] = all[i];
                tmp[1] = all[i+1];
                outlierlist[n][0] = parseShort(tmp);
                tmp[0] = all[i+2];
                tmp[1] = all[i+3];
                outlierlist[n][1] = parseShort(tmp);
                n++;
            }
            
            
        } catch (IOException e) {
            return outlierlist;
        }
        return outlierlist;
    }
    /** Swap the bytes of the provided integer in the reverse order. */


    private int swap(int i) {
	int j = i>>>24;
	j |= ((i >>> 16) & 0xff) <<8;
	j |= ((i >>> 8) & 0xff) << 16;
	j |= (i & 0xff) << 24;
	return j;
    }

    /** Swap the bytes of the provided short integer in the reverse order. */


    short swap(short i) {
	return (short)(((i >>> 8) & 0xff) | ((i & 0xff) << 8));
    }

    /** Parse a little-endian short integer from a byte array. */

    short parseShort(byte[] buf) {
	return (short)((buf[0] & 0xff) | ((buf[1] & 0xff) << 8));
    }

    /** turn long into a byte array **/
    private byte[] int2byte(long i){
        byte [] b = new byte[4];
        if(littleEndian){
            b[0] = (byte)(i & 0xff);
            i = i >> 8;
            b[1] = (byte)(i & 0xff);
            i = i >> 8;
            b[2] = (byte)(i & 0xff);
            i = i >> 8;
            b[3] = (byte)(i & 0xff);
        } else{
            b[3] = (byte)(i & 0xff);
            i = i >> 8;
            b[2] = (byte)(i & 0xff);
            i = i >> 8;
            b[1] = (byte)(i & 0xff);
            i = i >> 8;
            b[0] = (byte)(i & 0xff);
        }
        return b;
    }
    /** turn short into a byte array **/
    private byte[] short2byte(short[] s){
        byte [] b = new byte[4];
        int i;
        if(littleEndian){
            b[0] = (byte)(s[0] & 0xff);
            i = s[0] >> 8;
            b[1] = (byte)(i & 0xff);
            
            b[2] = (byte)(s[1] & 0xff);
            i = s[1] >> 8;
            b[3] = (byte)(i & 0xff);
        } else{
            b[1] = (byte)(s[0] & 0xff);
            i = s[0] >> 8;
            b[0] = (byte)(i & 0xff);
            
            b[3] = (byte)(s[1]& 0xff);
            i = s[1] >> 8;
            b[2] = (byte)(i & 0xff);
        }
        return b;
    }
    
    public float[][] getIntensities(){
        return intensities;
    }
    
    public int getWidth(){
        return nRows;
    }
    public int getHeight(){
        return nColumns;
    }
    
    public int[] getPixStats(){
        
        int size = nCells;
        int step = (int)(size/100);
        float[] pixels = new float[size];
        int[] quantiles = new int[101];
        for( int j =0; j < nColumns; j++){
            for( int i =0; i < nRows; i++){
                pixels[j*nRows+i] = intensities[i][j];
            }
                
        }
        java.util.Arrays.sort(pixels);
        quantiles[0] = (int) pixels[0];
        for (int i =1; i <  100; i++){
            quantiles[i] = (int) pixels[i*step-1];
        }
        quantiles[100] = (int) pixels[size-1];
        
        return quantiles;
    }
    
    public void getHistogram(short[][] outlierlist,  String headerinfo){
        int size = nCells;
        int step = (int)(size/100);
        double[] pixels = new double[size];
        double[] outlierpixels = new double[outlierlist.length];
        double[] notoutlierpixels = new double[size - outlierlist.length];
        int[] quantiles = new int[101];
        boolean[][] isOutlier = new boolean[nRows][nColumns];
        
        System.out.println("nColumns" + nColumns);
        System.out.println("size" + size);
        for( int j =0; j < nColumns; j++){
            for( int i =0; i < nRows; i++){
                pixels[j*nRows+i] = intensities[i][j];
            }
        }
        java.util.Arrays.sort(pixels);
        
        for (int i = 0; i < outlierlist.length; i ++){
            outlierpixels[i] = intensities[outlierlist[i][0]][outlierlist[i][1]];
            isOutlier[outlierlist[i][0]][outlierlist[i][1]] = true;
        }
        System.out.println("nOutlierlist" + outlierlist.length);
        int k = 0;
        for( int j =0; j < nColumns; j++){
            for( int i =0; i < nRows; i++){
                if (! isOutlier[i][j]) {
                    notoutlierpixels[k] = intensities[i][j];
                    k++;
                }
            }
        }
        
        String out = "";
        String out2 = "";
        double idealBinSize = (pixels[(int)(.99*size)]/100);
        double MIN = pixels[0];
        double MAX = pixels[size-1];        
        double[][] data = new double[3][];
        data[2] = pixels;
        data[0] = outlierpixels;
        data[1] = notoutlierpixels;
        double area = Math.round(outlierlist.length*1000.0/nCells);
        area = area / 10.0;
        //System.out.println("Number of bins = " + (int)((MAX-MIN)/idealBinSize)); 
        //System.out.println("MIN= " + MIN);
        //System.out.println("MAX= " + MAX);
        //System.out.println("Ideal binsize= " + idealBinSize);
        HistogramPlot.showHistograms(data, Math.min(1000, (int)((MAX-MIN)/idealBinSize)), headerinfo + 
                "\n(Min,Max) intensities = ("+MIN + ", "+MAX +")"+ 
                "\nOutlier size: " +area +"%");

    }
    public int[] createHistogram(float MIN, double MAX, double[] array, int binsize){
        double step = (MAX-MIN)/binsize;
        int[] hist = new int[binsize];
        int L = array.length;
        for (int i = 0 ; i < L; i ++){
            int pos = (int)((array[i] - MIN)/step);
            if (pos > binsize-1) continue;
            hist[pos] ++;
        }
        return hist;
    }
    public int getMean(){
        float ave = 0.0f;
        for( int j =0; j < nColumns; j++){
            for( int i =0; i < nRows; i++){
                ave += intensities[i][j];
            }
        }
        return (int)(ave/(nColumns*nRows));
    }

    public float getMaxIntensity(){
        return  maxIntensity;
    }
    public static void main (String[] args){ 
        
    }

}
